"""
Functionality to apply a mass dependent correction to quantities that have been
binned in mass bins (e.g. a mass function).
"""

import numpy as np
import yaml
import os
import scipy.interpolate as interpol
import unyt
from typing import Tuple


class VelociraptorBoxSizeCorrection:
    def __init__(self, filename: str, correction_directory: str):
        correction_file = f"{correction_directory}/{filename}"
        if not os.path.exists(correction_file):
            raise FileNotFoundError(f"Could not find {correction_file}!")
        with open(correction_file, "r") as handle:
            correction_data = yaml.safe_load(handle)
        self.is_log_x = correction_data["is_log_x"]
        self.x_min, self.x_max = correction_data["x_limits"]
        x = np.array(correction_data["x"])
        y = np.array(correction_data["y"])
        self.correction_spline = interpol.InterpolatedUnivariateSpline(x, y)

    def apply_mass_function_correction(
        self,
        mass_function_output: Tuple[unyt.unyt_array, unyt.unyt_array, unyt.unyt_array],
    ) -> Tuple[unyt.unyt_array, unyt.unyt_array, unyt.unyt_array]:

        bin_centers, mass_function, error = mass_function_output

        x_vals = bin_centers
        correction = np.ones(x_vals.shape)
        if self.is_log_x:
            x_vals = np.log10(x_vals)
        # only apply the correction to bins that are within the range for which
        # the correction is valid
        x_mask = (self.x_min <= x_vals) & (x_vals <= self.x_max)
        correction[x_mask] = self.correction_spline(x_vals[x_mask])

        corrected_mass_function = mass_function * correction
        corrected_mass_function.name = mass_function.name

        return bin_centers, corrected_mass_function, error



"""
All classes that are involved in the handling of the units, including
the master VelociraptorUnits class.
"""


import unyt
import h5py

from astropy.cosmology.core import Cosmology
from astropy.cosmology import wCDM, FlatLambdaCDM


class VelociraptorUnits(object):
    """
    Generates a unyt system that can then be used with the velociraptor data.

    You are probably looking for the following attributes:

    + VelociraptorUnits.length
    + VelociraptorUnits.mass
    + VelociraptorUnits.metallicity (relative to solar)
    + VelociraptorUnits.age
    + VelociraptorUnits.velocity
    + VelociraptorUnits.star_formation_rate

    This will allow you to extract variables in the correct units. This object
    also holds the current scale factor and redshift through the a and z variables.
    Finally, it contains whether or not the current unit system is comoving
    (VelociraptorUnits.comoving) and whether or not the underlying simulation
    was cosmological (VelociraptorUnits.cosmological).
    """

    # Pre-define these for autocomplete.
    length: unyt.unyt_quantity
    mass: unyt.unyt_quantity
    metallicity: unyt.unyt_quantity
    age: unyt.unyt_quantity
    velocity: unyt.unyt_quantity
    star_formation_rate: unyt.unyt_quantity
    period: unyt.unyt_quantity
    box_volume: unyt.unyt_quantity
    cosmology: Cosmology

    def __init__(self, filename: str, disregard_units: bool = False):
        """
        Creates an instance of the ``VelociraptorUnits`` Class.

        Parameters
        ----------

        filename: str
            Filename of the VELOCIraptor catalogue (i.e. .properties)
            file that contains the unit metadata in its header.

        disregard_units: bool, optional
            If ``True``, then disregard any additional units in the
            VELOCIraptor catalogues, and instead base everything on
            the 'base' units of velocity, length, and mass. In this
            case metallicities are left dimensionless. If you are
            using EAGLE data, you should set this to False, as the
            star formation rate units are presented in non-internal
            units.

        """

        self.filename = filename
        self.disregard_units = disregard_units

        self.get_unit_dictionary()

        return

    def get_unit_dictionary(self):
        """
        Gets the unit library from the header information in the file.
        These are a mix of units, so we just read them all -- and allow the
        people who define the registration functions to figure out how to
        use them.
        """

        self.units = {}

        with h5py.File(self.filename, "r") as handle:
            attributes = handle.attrs

            self.units["length"] = attributes["Length_unit_to_kpc"] * unyt.kpc
            self.units["mass"] = attributes["Mass_unit_to_solarmass"] * unyt.Solar_Mass
            self.units["velocity"] = attributes["Velocity_to_kms"] * unyt.km / unyt.s

            metallicity_units = unyt.dimensionless
            stellar_age_units = (self.units["length"] / self.units["velocity"]).to(
                "Year"
            )
            star_formation_units = (
                self.units["velocity"] * self.units["mass"] / self.units["length"]
            ).to("Solar_Mass / Year")

            if not self.disregard_units:
                # Extract units that may not be present in the file
                try:
                    metallicity_units = (
                        attributes["Metallicity_unit_to_solar"] * unyt.dimensionless
                    )
                except (AttributeError, KeyError):
                    pass

                try:
                    stellar_age_units = attributes["Stellar_age_unit_to_yr"] * unyt.year
                except (AttributeError, KeyError):
                    pass

                try:
                    star_formation_units = (
                        attributes["SFR_unit_to_solarmassperyear"]
                        * unyt.Solar_Mass
                        / unyt.year
                    )
                except (AttributeError, KeyError):
                    pass

            self.units["metallicity"] = metallicity_units
            self.units["age"] = stellar_age_units
            self.units["star_formation_rate"] = star_formation_units

            self.scale_factor = attributes["Time"]
            self.a = self.scale_factor
            self.redshift = 1.0 / self.a - 1.0
            self.z = self.redshift

            self.cosmological = bool(attributes["Cosmological_Sim"])
            self.comoving = bool(attributes["Comoving_or_Physical"])
            self.cosmology = self.load_cosmology(handle)

            # Period is comoving.
            self.period = unyt.unyt_quantity(
                attributes["Period"], units=self.units["length"]
            )
            self.box_length = unyt.unyt_quantity(
                attributes["Period"] / self.a, units=self.units["length"]
            )
            self.comoving_box_volume = self.box_length ** 3
            self.physical_box_volume = self.period ** 3

        # Unpack the dictionary to variables
        for name, unit in self.units.items():
            setattr(self, name, unit)

        return

    def load_cosmology(self, handle: h5py.File) -> Cosmology:
        """
        Save the (astropy) cosmology to a HDF5 dataset.

        Parameters
        ----------

        handle: h5py.File
            h5py file handle for the catalogue file.

        Returns
        -------

        astropy.cosmology.Cosmology:
            Astropy cosmology instance extracted from the HDF5 file.
            Also sets ``self.cosmology``.
        
        """

        try:
            group = handle["Configuration"].attrs
        except:
            return None

        # Note that some of these are unused - commented out if not.
        H0 = 100.0 * float(group["h_val"])
        w_of_DE = float(group["w_of_DE"])
        Omega_DE = float(group["Omega_DE"])
        # Omega_Lambda = float(group["Omega_Lambda"])
        Omega_b = float(group["Omega_b"])
        # Omega_cdm = float(group["Omega_cdm"])
        # Omega_k = float(group["Omega_k"])
        Omega_m = float(group["Omega_m"])
        # Omega_nu = float(group["Omega_nu"])
        # Omega_r = float(group["Omega_r"])

        if w_of_DE != -1.0:
            cosmology = wCDM(H0=H0, Om0=Omega_m, Ode0=Omega_DE, w0=w_of_DE, Ob0=Omega_b)
        else:
            # No EoS
            cosmology = FlatLambdaCDM(H0=H0, Om0=Omega_m, Ob0=Omega_b)

        self.cosmology = cosmology

        return cosmology



"""
Default registration functions.

If you add one, don't forget to add it to global_registration_functions
at the end of the file.
"""

import unyt

from typing import Union

from velociraptor.exceptions import RegistrationDoesNotMatchError
from velociraptor.units import VelociraptorUnits
from velociraptor.regex import cached_regex
from velociraptor.catalogue.translator import (
    get_aperture_unit,
    get_particle_property_name_conversion,
)


def registration_fail_all(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Basic registration function showing function signature that is
    required and automatically fails all tests against itself.

    Function signature:

    + field_path: the name of the field
    + unit_system: a VelociraptorUnits instance that contains all unit
      information that is available from the velociraptor catalogue

    Return signature:

    + field_units: the units that correspond to field_path.
    + name: A fancy (possibly LaTeX'd) name for the field.
    + snake_case: A correct snake_case name for the field.
    """

    if field_path == "ThisFieldPathWouldNeverExist":
        return (
            unit_system.length,
            "Fancy $N_{\\rm ever}$ exists",
            "this_field_path_would_never_exist",
        )
    else:
        raise RegistrationDoesNotMatchError


def registration_apertures(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Registers aperture values by searching them with regex.
    """

    # Capture group 1: quantity
    # Capture group 2: particle type
    # Capture group 3: sf / nsf
    # Capture group 4: size of aperture

    match_string = "Aperture_([^_]*)_([a-zA-Z]*)?_?([a-zA-Z]*)?_?([0-9]*)_kpc"
    regex = cached_regex(match_string)

    match = regex.match(field_path)

    if match:
        quantity = match.group(1)
        ptype = match.group(2)
        star_forming = match.group(3)
        aperture_size = int(match.group(4))

        unit = get_aperture_unit(quantity, unit_system)
        name = get_particle_property_name_conversion(quantity, ptype)

        if star_forming:
            sf_in_name = f"{star_forming.upper()} "
        else:
            sf_in_name = ""

        full_name = f"{sf_in_name}{name} ({aperture_size} kpc)"
        snake_case = field_path.lower().replace("aperture_", "")

        return unit, full_name, snake_case
    else:
        raise RegistrationDoesNotMatchError


def registration_projected_apertures(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Registers aperture values by searching them with regex.
    """

    # Capture group 1: aperture number
    # Capture group 2: quantity
    # Capture group 3: particle type
    # Capture group 4: sf / nsf
    # Capture group 5: size of aperture

    match_string = (
        "Projected_aperture_([0-9])_([^_]*)_([a-zA-Z]*)?_?([a-zA-Z]*)?_?([0-9]*)_kpc"
    )
    regex = cached_regex(match_string)

    match = regex.match(field_path)

    if match:
        aperture = match.group(1)
        quantity = match.group(2)
        ptype = match.group(3)
        star_forming = match.group(4)
        aperture_size = int(match.group(5))

        unit = get_aperture_unit(quantity, unit_system)
        name = get_particle_property_name_conversion(quantity, ptype)

        if star_forming:
            sf_in_name = f"{star_forming.upper()} "
        else:
            sf_in_name = ""

        full_name = f"{sf_in_name}{name} (Proj. {aperture}, {aperture_size} kpc)"
        snake_case = field_path.lower().replace("aperture_", "")

        return unit, full_name, snake_case
    else:
        raise RegistrationDoesNotMatchError


def registration_energies(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Registers all energy related quantities (those beginning with E).
    """

    if not field_path[:2] in ["Ef", "Ek", "Ep", "En"]:
        raise RegistrationDoesNotMatchError

    if field_path[:5] == "Efrac":
        # This is an energy _fraction_
        full_name = "Energy Fraction"
        unit = unyt.dimensionless
    else:
        # This is an absolute energy
        if field_path[:4] == "Ekin":
            full_name = "Kinetic Energy"
        elif field_path[:4] == "Epot":
            full_name = "Potential Energy"
        else:
            full_name = "Energy"

        unit = unit_system.mass * unit_system.velocity * unit_system.velocity

    return unit, full_name, field_path.lower()


def registration_ids(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Registers all quantities related to particle ids and halo ids (those beginning or ending with ID).
    """

    if not (field_path[:2] == "ID" or field_path[-2:] == "ID"):
        raise RegistrationDoesNotMatchError

    # As identifiers, all of these quantities are dimensionless
    unit = unyt.dimensionless

    if field_path == "ID":
        full_name = "Halo ID"
    elif field_path == "ID_mpb":
        full_name = "ID of Most Bound Particle"
    elif field_path == "ID_minpot":
        full_name = "ID of Particle at Potential Minimum"
    elif field_path == "hostHaloID":
        full_name = "Host Halo ID"
    else:
        full_name = "Generic ID"

    return unit, full_name, field_path.lower()


def registration_rotational_support(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Registers rotational support quantities (those beginning with K).
    Note that this corresponds to \kappa in Sales+2010 _not_ K.
    """

    if not field_path[0] == "K":
        raise RegistrationDoesNotMatchError

    # All quantities are ratios and so are dimensionless
    unit = unyt.dimensionless

    # Capture group 1: particle type
    # Capture group 2: star forming / not star forming

    match_string = "Krot_?([a-z]*)_?([a-z]*)?"
    regex = cached_regex(match_string)

    match = regex.match(field_path)

    if match:
        ptype = match.group(1)
        star_forming = match.group(2)

        full_name = "$\\kappa_{{\\rm rot}"

        if ptype:
            full_name += f", {{\\rm {ptype}}}"

        full_name += "}$"

        if star_forming:
            full_name += f" ({star_forming.upper()})"
    else:
        raise RegistrationDoesNotMatchError

    return unit, full_name, field_path.lower()


def registration_angular_momentum(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Registers values starting with L, those that represent
    angular momenta.
    """

    if not field_path[0] == "L":
        raise RegistrationDoesNotMatchError

    # All are angular momenta, so have same units.
    unit = unit_system.mass * unit_system.length * unit_system.velocity

    # Capture group 1: axis (x, y, z)
    # Capture group 2: radius within this was calculated, e.g. 200crit
    # Capture group 3: excluding or not excluding
    # Capture group 4: particle type
    # Capture group 5: star forming?

    match_string = "L([a-z])_?([A-Z]*[0-9]+[a-z]*)?_?(excl)?_?([a-z]*)_?([a-z]*)"
    regex = cached_regex(match_string)
    match = regex.match(field_path)

    if match:
        axis = match.group(1)
        radius = match.group(2)
        excluding = match.group(3)
        ptype = match.group(4)
        star_forming = match.group(5)

        full_name = "$L_{"

        if axis:
            full_name += axis
        if radius:
            full_name += f", {{\\rm {radius}}}"

        full_name += "}$"

        if ptype:
            full_name += " ("

            if excluding:
                full_name += "Excl. "

            cap_ptype = ptype[0].upper() + ptype[1:]

            full_name += cap_ptype

            if star_forming:
                full_name += f", {star_forming.upper()}"

            full_name += ")"
    else:
        raise RegistrationDoesNotMatchError

    return unit, full_name, field_path.lower()


def registration_masses(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Registration for all mass-based quantities. (Start with M)
    """

    if not field_path[0] == "M":
        raise RegistrationDoesNotMatchError

    # All, obviously, have a unit of mass
    unit = unit_system.mass
    full_name = ""

    # Deal with special cases.
    if field_path == "Mvir":
        full_name = "$M_{\\rm vir}$"
    elif field_path == "Mass_FOF":
        full_name = "$M_{\\rm FOF}$"
    elif field_path == "Mass_tot":
        full_name = r"$M$"
    elif field_path == "Mass_interloper":
        full_name = "$M_{\\rm BG}$"

    # General regex matching case.

    # Capture group 1: Mass or M
    # Capture group 2: radius within this was calculated, e.g. 200crit
    # Capture group 3: excluding?
    # Capture group 4: ptype
    # Capture group 5: star forming?
    # Capture group 6: "other"
    match_string = (
        "(Mass|M)_?([A-Z]*[0-9]+[a-z]*)?_?(excl)?_?([a-z]*)_?(nsf|sf)?_?([a-zA-Z0-9]*)"
    )
    regex = cached_regex(match_string)
    match = regex.match(field_path)
    if match and not full_name:
        mass = match.group(1)
        radius = match.group(2)
        excluding = match.group(3)
        ptype = match.group(4)
        star_forming = match.group(5)
        other = match.group(6)

        full_name = "$M"

        if radius:
            full_name += f"_{{\\rm {radius}}}"
        elif other:
            full_name += f"_{{\\rm {other}}}"

        full_name += "$"

        if ptype:
            full_name += " ("

            if excluding:
                full_name += "Excl. "

            cap_ptype = ptype[0].upper() + ptype[1:]

            full_name += cap_ptype

            if star_forming:
                full_name += f", {star_forming.upper()}"

            full_name += ")"

    return unit, full_name, field_path.lower()


def registration_rvmax_quantities(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Registration for all quantities measured within RVmax (Start with RVmax)
    """

    if not field_path[:5] == "RVmax":
        raise RegistrationDoesNotMatchError

    # Capture group 1: Eigenvector or velocity dispersion
    # Capture group 2: xx, xy, etc. for above
    # Capture group 3: Angular momentum quantity
    # Capture group 4: x, y, z for angular momentum
    # Capture group 5: catch all others
    match_string = "RVmax_((eig|veldisp)_([a-z]{2}))?_?(L([a-z]))?_?([a-zA-Z0-9_]*)"
    regex = cached_regex(match_string)
    match = regex.match(field_path)

    if match:
        raise RegistrationDoesNotMatchError

    return  # TODO


def registration_radii(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Registration for all radii quantities (start with R_)
    """

    # First, two special cases.
    if field_path == "Rvir":
        full_name = "$R_{\\rm vir}$"
    elif field_path == "Rmax":
        full_name = "$R_{\\rm max}$"
    elif field_path[:2] != "R_":
        raise RegistrationDoesNotMatchError

    unit = unit_system.length

    # Capture group 1: Characteristic scale
    # Capture group 2: Excluding?
    # Capture group 3: particle type
    # Capture group 4: star forming?
    match_string = "R_([a-zA-Z0-9]*)_?(excl)?_?([a-z]*)?_?(sf|nsf)?"
    regex = cached_regex(match_string)
    match = regex.match(field_path)

    if match:
        radius = match.group(1)
        excluding = match.group(2)
        ptype = match.group(3)
        star_forming = match.group(4)

        full_name = "$R"

        if radius:
            full_name += f"_{{\\rm {radius}}}"

        full_name += "$"

        if ptype:
            full_name += " ("

            if excluding:
                full_name += "Excl. "

            cap_ptype = ptype[0].upper() + ptype[1:]

            full_name += cap_ptype

            if star_forming:
                full_name += f", {star_forming.upper()}"

            full_name += ")"

    return unit, full_name, field_path.lower()


def registration_star_formation_rate(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Registers star formation rate quantities. (Start with SFR)
    """

    if not field_path[:3] == "SFR":
        raise RegistrationDoesNotMatchError

    unit = unit_system.star_formation_rate

    full_name = r"Star Formation Rate $\dot{\rho}_*$"

    return unit, full_name, field_path.lower()


def registration_temperature(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Registers temperature based quantites (Those beginning with T).
    """

    if not field_path[0] == "T":
        raise RegistrationDoesNotMatchError

    unit = unyt.K

    # Capture group 1: particle type
    # Capture group 2: star forming?
    match_string = "T_?([a-z]*)?_?(sf|nsf)?"
    regex = cached_regex(match_string)
    match = regex.match(field_path)

    if match:
        ptype = match.group(1)
        star_forming = match.group(2)

        full_name = "$T$"

        if ptype:
            full_name += " ("

            cap_ptype = ptype[0].upper() + ptype[1:]

            full_name += cap_ptype

            if star_forming:
                full_name += f", {star_forming.upper()}"

            full_name += ")"

    return unit, full_name, field_path.lower()


def registration_structure_type(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Registers the StructureType field.
    """

    if not field_path == "Structuretype":
        raise RegistrationDoesNotMatchError

    return unyt.dimensionless, "Structure Type", field_path.lower()


def registration_velocities(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Registers velocity quantities (those starting with V).
    """

    if not field_path[0] == "V":
        raise RegistrationDoesNotMatchError

    unit = unit_system.velocity

    if field_path == "Vmax":
        # Special case, handle here
        full_name = "$V_{\\rm max}$"
    else:
        # Need to do a regex search
        # Capture group 1: X, Y, Z
        # Capture group 2: mbp or minpot? Could be empty.
        # Capture group 4: gas/star.
        match_string = "V(X|Y|Z)c([a-z]*)?(_([a-z]*))?"
        regex = cached_regex(match_string)
        match = regex.match(field_path)

        if match:
            coordinate = match.group(1)
            ptype = match.group(4)
            misc = match.group(2)

            full_name = f"$V_{coordinate.lower()}$"

            if ptype:
                full_name += f" ({ptype})"

            if misc:
                if misc == "mbp":
                    full_name = f"Most bound particle {full_name}"
                elif misc == "minpot":
                    full_name = f"Minimum potential {full_name}"
                else:
                    full_name = f"{misc} {full_name}"
            else:
                full_name = "CoM " + full_name
        else:
            raise RegistrationDoesNotMatchError

    return unit, full_name, field_path.lower()


def registration_positions(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Registers all positon based quantities (those beginning with X, Y, or Z).
    """

    if not field_path[0] in ["X", "Y", "Z"] and not field_path[:4] == "Zmet":
        raise RegistrationDoesNotMatchError

    # All position quantities have units of length
    unit = unit_system.length

    # Capture group 1: x, y, or z
    # Capture group 2: ignore
    # Capture group 3: ptype
    # Capture group 4: misc info, e.g. mbp or minpot
    match_string = "(X|Y|Z)c(_([a-z]*))?([a-z]*)?"
    regex = cached_regex(match_string)
    match = regex.match(field_path)

    if match:
        coordinate = match.group(1)
        ptype = match.group(3)
        misc = match.group(4)

        full_name = f"${coordinate.lower()}$"

        if ptype:
            full_name += f" ({ptype})"

        if misc:
            if misc == "mbp":
                full_name = f"Most bound particle {full_name}"
            elif misc == "minpot":
                full_name = f"Minimum potential {full_name}"
            else:
                full_name = f"{misc} {full_name}"
        else:
            full_name = "CoM " + full_name
    else:
        raise RegistrationDoesNotMatchError

    return unit, full_name, field_path.lower()


def registration_concentration(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Registers concentration values (those beginning with c).
    """

    if not field_path == "cNFW":
        raise RegistrationDoesNotMatchError

    return unyt.dimensionless, "Concentration $c_{\\rm NFW}$", field_path.lower()


def registration_metallicity(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Registers metallicity-based quantities (those beginning with Zmet)
    """

    if not field_path[:4] == "Zmet":
        raise RegistrationDoesNotMatchError

    unit = unit_system.metallicity

    # Need to do a regex search
    # Capture group 1: gas/star.
    # Capture group 2: star forming?
    match_string = "Zmet_?([a-z]*)_?(sf|nsf)?"
    regex = cached_regex(match_string)
    match = regex.match(field_path)

    if match:
        ptype = match.group(1)
        star_forming = match.group(2)

        full_name = "Metallicity $Z"

        if ptype:
            if ptype == "gas":
                full_name += "_g"
            elif ptype == "star":
                full_name += "_*"

            cap_ptype = ptype[0].upper() + ptype[1:]
            full_name = f"{cap_ptype} {full_name}"

        full_name += "$"

        if star_forming:
            full_name += f" ({star_forming.upper()})"
    else:
        raise RegistrationDoesNotMatchError

    return unit, full_name, field_path.lower()


def registration_eigenvectors(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Registers eigenvector quantities (those beginning with eig).
    """

    if not field_path[:3] == "eig":
        raise RegistrationDoesNotMatchError

    unit = unit_system.length

    # Need to do a regex search
    # Capture group 1: xy, etc.
    # Capture group 2: gas/star.
    match_string = "eig_([a-z][a-z])_?([a-z]*)?"
    regex = cached_regex(match_string)
    match = regex.match(field_path)

    if match:
        coordinate = match.group(1)
        ptype = match.group(2)

        full_name = f"$\\hat{{r}}_{{{{\\rm v}}, {coordinate.lower()}}}$"

        if ptype:
            full_name += f" ({ptype})"
    else:
        raise RegistrationDoesNotMatchError

    return unit, full_name, field_path.lower()


def registration_veldisp(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Registers velocity dispersion quantities (those beginning with veldisp).
    """

    if not field_path[:7] == "veldisp":
        raise RegistrationDoesNotMatchError

    unit = unit_system.velocity

    # Need to do a regex search
    # Capture group 1: xy, etc.
    # Capture group 2: gas/star.
    match_string = "veldisp_([a-z][a-z])_?([a-z]*)?"
    regex = cached_regex(match_string)
    match = regex.match(field_path)

    if match:
        coordinate = match.group(1)
        ptype = match.group(2)

        full_name = f"$\sigma_{{{{\\rm v}}, {coordinate.lower()}}}$"

        if ptype:
            full_name += f" ({ptype})"
    else:
        raise RegistrationDoesNotMatchError

    return unit, full_name, field_path.lower()


def registration_stellar_age(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Registers the stellar ages properties (currently tage_star).
    """

    if field_path == "tage_star":
        return unit_system.age, "Mean Stellar Age", field_path.lower()
    else:
        raise RegistrationDoesNotMatchError


def registration_element_mass_fractions(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Registers the element mass fraction properties.

    Hopefully this is changed in the future as this is a mess.
    """

    if not field_path[:20] == "ElementMassFractions":
        raise RegistrationDoesNotMatchError

    unit = unit_system.metallicity

    # Need to do a regex search
    # Capture group 1,2: index number - if not present default to 0
    # Capture group 3: mass weighted?
    # Capture group 4: units
    # Capture group 5: particle typr
    match_string = (
        "ElementMassFractions(_index_)?([0-9]*)_([a-zA-Z]+)_([a-zA-Z]+)_?([a-z]*)"
    )
    regex = cached_regex(match_string)
    match = regex.match(field_path)

    snake_case = "element"

    if match:
        index = match.group(2) if match.group(2) else 0
        mass_weighted = match.group(3)
        extracted_units = match.group(4)
        ptype = match.group(5)

        full_name = f"Element {index} Mass Fraction"
        snake_case = f"{snake_case}_{index}"

        if ptype:
            cap_ptype = ptype[0].upper() + ptype[1:]
            full_name = f"{cap_ptype} {full_name}"

            snake_case += f"_{ptype}"

    else:
        raise RegistrationDoesNotMatchError

    return unit, full_name, snake_case


def registration_dust_mass_fractions(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Registers the dust mass fraction properties.

    Hopefully this is changed in the future as this is a mess.
    """

    if not field_path[:17] == "DustMassFractions":
        raise RegistrationDoesNotMatchError

    unit = unit_system.metallicity

    # Need to do a regex search
    # Capture group 1,2: index number - if not present default to 0
    # Capture group 3: mass weighted?
    # Capture group 4: units
    # Capture group 5: particle typr
    match_string = (
        "DustMassFractions(_index_)?([0-9]*)_([a-zA-Z]+)_([a-zA-Z]+)_?([a-z]*)"
    )
    regex = cached_regex(match_string)
    match = regex.match(field_path)

    snake_case = "dust"
    if match:
        index = match.group(2) if match.group(2) else 0
        mass_weighted = match.group(3)
        extracted_units = match.group(4)
        ptype = match.group(5)

        full_name = f"Dust {index} Mass Fraction"
        snake_case = f"{snake_case}_{index}"

        if ptype:
            cap_ptype = ptype[0].upper() + ptype[1:]
            full_name = f"{cap_ptype} {full_name}"

            snake_case += f"_{ptype}"

    else:
        raise RegistrationDoesNotMatchError

    return unit, full_name, snake_case


def registration_number(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Registers the number of particles in each halo (n_{bh, gas, star} and npart).
    """

    if field_path[:2] == "n_":
        unit = unyt.dimensionless
        switch = {
            "bh": "Black Hole",
            "gas": "Gas",
            "star": "Star",
            "interloper": "Interloper",
        }
        snake_case = field_path[2:]
        full_name = f"Number of {switch.get(snake_case, 'Unknown')} Particles"

    elif field_path == "npart":
        unit = unyt.dimensionless
        full_name = "Number of Particles"
        snake_case = "part"

    else:
        raise RegistrationDoesNotMatchError

    return unit, full_name, snake_case


def registration_gas_H_and_He_masses(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Registers the masses in Hydrogen & Helium within apertures
    """

    unit = unit_system.mass

    # Capture aperture size
    match_string = "Aperture_([a-zA-Z]*)_aperture_total_gas_([0-9]*)_kpc"
    regex = cached_regex(match_string)

    match = regex.match(field_path)

    if match:
        long_species = match.group(1)
        aperture_size = match.group(2)

        try:
            short_species = {"HeliumMasses": "He", "HydrogenMasses": "H"}[long_species]
            long_name_species = {
                "HeliumMasses": "Helium",
                "HydrogenMasses": "Hydrogen",
            }[long_species]
            math_name = {"HeliumMasses": "M_{\\rm He}", "HydrogenMasses": "M_{\\rm H}"}[
                long_species
            ]
        except KeyError:
            raise RegistrationDoesNotMatchError

        full_name = f"{long_name_species} Gas Mass {math_name} ({aperture_size} kpc)"
        snake_case = f"{short_species}_mass_{aperture_size}_kpc"

        return unit, full_name, snake_case
    else:
        raise RegistrationDoesNotMatchError


def registration_gas_diffuse_element_masses(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Registers the masses in Hydrogen & Helium within apertures
    """

    unit = unit_system.mass

    # Capture aperture size
    match_string = "Aperture_Diffuse([a-zA-Z]*)MassesFrom(Table|Model)_aperture_total_gas_([0-9]*)_kpc"
    regex = cached_regex(match_string)

    match = regex.match(field_path)

    if match:
        element = match.group(1)
        table_model = match.group(2)
        aperture_size = match.group(3)
        try:
            long_name_element = (
                f"Diffuse {element} from depletion {table_model.lower()}"
            )
            element_symbol = {
                "Carbon": "C",
                "Oxygen": "O",
                "Magnesium": "Mg",
                "Silicon": "Si",
                "Iron": "Fe",
            }[element]
            math_name = (
                f"M^{{\\rm {table_model.lower()}}}_{{\\rm {element_symbol}, diffuse}}"
            )
        except KeyError:
            raise RegistrationDoesNotMatchError

        full_name = (
            f"{element} ({table_model}) Gas Mass {element_symbol} ({aperture_size} kpc)"
        )
        snake_case = f"{element.lower()}_mass_{table_model.lower()}_{aperture_size}_kpc"

        return unit, full_name, snake_case
    else:
        raise RegistrationDoesNotMatchError


def registration_dust_masses_from_table(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Registers the dust mass fraction properties.
    """

    if not field_path[:28] == "Aperture_DustMassesFromTable":
        raise RegistrationDoesNotMatchError

    unit = unit_system.mass

    match_string = "Aperture_DustMassesFromTable_aperture_total_gas_([0-9]*)_kpc"
    regex = cached_regex(match_string)
    match = regex.match(field_path)

    if match:
        aperture_size = match.group(1)
        full_name = f"Total Dust Mass from Tables ({aperture_size} kpc)"
        snake_case = f"dust_mass_table_{aperture_size}_kpc"
    else:
        raise RegistrationDoesNotMatchError

    return unit, full_name, snake_case


def registration_gas_hydrogen_species_masses(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Registers the masses in hydrogen species within apertures.
    """

    unit = unit_system.mass

    # Capture aperture size
    match_string = "Aperture_([a-zA-Z]*)_(index_0_)?aperture_total_gas_([0-9]*)_kpc"
    regex = cached_regex(match_string)

    match = regex.match(field_path)

    if match:
        long_species = match.group(1)
        aperture_size = match.group(3)

        try:
            short_species = {
                "AtomicHydrogenMasses": "HI",
                "IonisedHydrogenMasses": "HII",
                "MolecularHydrogenMasses": "H2",
            }[long_species]
            full_name_species = {
                "AtomicHydrogenMasses": "HI",
                "IonisedHydrogenMasses": "HII",
                "MolecularHydrogenMasses": "H$_2$",
            }[long_species]
            math_name_species = {
                "AtomicHydrogenMasses": "$M_{\\rm HI}$",
                "IonisedHydrogenMasses": "$M_{\\rm HII}$",
                "MolecularHydrogenMasses": "$M_{\\rm H_2}$",
            }[long_species]

        except KeyError:
            raise RegistrationDoesNotMatchError

        full_name = (
            f"{full_name_species} Gas Mass {math_name_species} ({aperture_size} kpc)"
        )
        snake_case = f"{short_species}_mass_{aperture_size}_kpc"

        return unit, full_name, snake_case
    else:
        raise RegistrationDoesNotMatchError


def registration_cold_dense_gas_properties(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Registers the mass of cold (T < 10^4.5 K), dense (nH > 0.1 cm^-3) gas in apertures.
    """
    unit = unit_system.mass

    # Capture aperture size
    match_string = "Aperture_ColdDense([a-zA-Z]*)Masses_aperture_total_gas_([0-9]*)_kpc"
    regex = cached_regex(match_string)
    match = regex.match(field_path)

    if match:
        quantity_key = match.group(1)
        aperture_size = match.group(2)

        try:
            long_quantity = {"DiffuseMetal": "Diffuse Metal", "Gas": "Gas"}[
                quantity_key
            ]
            short_quantity = {"DiffuseMetal": "diffuse_metal", "Gas": "gas"}[
                quantity_key
            ]
        except KeyError:
            raise RegistrationDoesNotMatchError
        full_name = (
            f"{long_quantity} Masses in Cold, Dense ($T < 10^{{4.5}}\;{{\rm K}}$, "
            f"$n_{{\\rm H}}$ > 0.1 \\; {{\rm cm^{{-3}}}}$) Gas ({aperture_size} kpc)"
        )
        snake_case = f"cold_dense_{short_quantity}_mass_{aperture_size}_kpc"
        return unit, full_name, snake_case
    else:
        raise RegistrationDoesNotMatchError


def registration_log_element_ratios_times_masses(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Registers the log10(Fe/H) times mass and log10(O/H) times mass within apertures for two particle floor values
    """

    unit = unit_system.mass

    # Capture aperture size
    match_string = (
        "Aperture_([a-zA-Z]*)Masses(Lo|Hi)Floor_aperture_total_([a-zA-Z]*)_([0-9]*)_kpc"
    )
    regex = cached_regex(match_string)

    match = regex.match(field_path)

    if match:
        long_species = match.group(1)
        floor_type = match.group(2)
        part_type = match.group(3)
        aperture_size = match.group(4)
        try:
            short_species = {
                "LogOxygenOverHydrogen": "O_over_H",
                "LogIronOverHydrogen": "Fe_over_H",
                "LogSNIaIronOverHydrogen": "SNIaFe_over_H",
                "LogOxygenOverHydrogenAtomic": "O_over_H_atomic",
                "LogOxygenOverHydrogenMolecular": "O_over_H_molecular",
            }[long_species]
            element_name = {
                "LogOxygenOverHydrogen": "Oxygen",
                "LogIronOverHydrogen": "Iron",
                "LogSNIaIronOverHydrogen": "SNIa Iron",
                "LogOxygenOverHydrogenAtomic": "Atomic-phase Oxygen",
                "LogOxygenOverHydrogenMolecular": "Molecular-phase Oxygen",
            }[long_species]
            fraction_name = {
                "LogOxygenOverHydrogen": "O/H",
                "LogIronOverHydrogen": "Fe/H",
                "LogSNIaIronOverHydrogen": "Fe_SNIa/H",
                "LogOxygenOverHydrogenAtomic": "O/H",
                "LogOxygenOverHydrogenMolecular": "O/H",
            }[long_species]

            short_floortype = {"Lo": "lowfloor", "Hi": "highfloor"}[floor_type]
            floor_value = {"Lo": "-4", "Hi": "-3"}[floor_type]
        except KeyError:
            raise RegistrationDoesNotMatchError

        full_name = (
            f"Log10 {element_name} Abundance Weighted {part_type.capitalize()} Mass ({fraction_name}) "
            f"$\times M_{{\\rm gas}}$, from particle values floored at [{fraction_name}]$\\gtreq {floor_value}$ "
            f"({aperture_size} kpc)"
        )
        snake_case = f"log_{short_species}_times_{part_type}_mass_{short_floortype}_{aperture_size}_kpc"
        return unit, full_name, snake_case
    else:
        raise RegistrationDoesNotMatchError


def registration_lin_element_ratios_times_masses(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Registers the Fe/H times mass and O/H times mass within apertures for two particle floor values
    """

    unit = unit_system.mass

    # Capture aperture size
    match_string = "Aperture_([a-zA-Z]*)Masses_aperture_total_([a-zA-Z]*)_([0-9]*)_kpc"
    regex = cached_regex(match_string)

    match = regex.match(field_path)

    if match:
        long_species = match.group(1)
        part_type = match.group(2)
        aperture_size = match.group(3)

        try:
            short_species = {
                "TotalOxygenOverHydrogen": "O_over_H_total",
                "OxygenOverHydrogen": "O_over_H",
                "IronOverHydrogen": "Fe_over_H",
            }[long_species]
            element_name = {
                "TotalOxygenOverHydrogen": "Oxygen",
                "OxygenOverHydrogen": "Oxygen",
                "IronOverHydrogen": "Iron",
            }[long_species]
            fraction_name = {
                "TotalOxygenOverHydrogen": "O/H",
                "OxygenOverHydrogen": "O/H",
                "IronOverHydrogen": "Fe/H",
            }[long_species]
        except KeyError:
            raise RegistrationDoesNotMatchError

        full_name = (
            f"Linear {element_name} Abundance Weighted {part_type.capitalize()} Mass ({fraction_name}) "
            f"$\times M_{{\\rm gas}}$ ({aperture_size} kpc)"
        )
        snake_case = f"lin_{short_species}_times_{part_type}_mass_{aperture_size}_kpc"
        return unit, full_name, snake_case
    else:
        raise RegistrationDoesNotMatchError


def registration_dust_masses(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Registers the masses in dust within apertures
    """

    unit = unit_system.mass

    # Capture aperture size
    match_string = "Aperture_([a-zA-Z]*)_aperture_total_gas_([0-9]*)_kpc"
    regex = cached_regex(match_string)

    match = regex.match(field_path)

    if match:
        long_species = match.group(1)
        aperture_size = match.group(2)

        try:
            short_species = {
                "GraphiteMasses": "graphite",
                "SilicatesMasses": "silicates",
                "SmallGrainMasses": "small_grain",
                "LargeGrainMasses": "large_grain",
                "GraphiteMassesAtomic": "atomic_graphite",
                "SilicatesMassesAtomic": "atomic_silicates",
                "SmallGrainMassesAtomic": "atomic_small_grain",
                "LargeGrainMassesAtomic": "atomic_large_grain",
                "GraphiteMassesMolecular": "molecular_graphite",
                "SilicatesMassesMolecular": "molecular_silicates",
                "SmallGrainMassesMolecular": "molecular_small_grain",
                "LargeGrainMassesMolecular": "molecular_large_grain",
                "GraphiteMassesColdDense": "cold_dense_graphite",
                "SilicatesMassesColdDense": "cold_dense_silicates",
                "SmallGrainMassesColdDense": "cold_dense_small_grain",
                "LargeGrainMassesColdDense": "cold_dense_large_grain",
            }[long_species]
            pretty_name = {
                "GraphiteMasses": "Graphite Dust Mass",
                "SilicatesMasses": "Silicate Dust Mass",
                "SmallGrainMasses": "small_grain",
                "LargeGrainMasses": "large_grain",
                "GraphiteMassesAtomic": "Graphite Dust Mass in Atomic Gas",
                "SilicatesMassesAtomic": "Silicate Dust Mass in Atomic Gas",
                "SmallGrainMassesAtomic": "atomic_small_grain",
                "LargeGrainMassesAtomic": "atomic_large_grain",
                "GraphiteMassesMolecular": "Graphite Dust Mass in Molecular Gas",
                "SilicatesMassesMolecular": "Silicate Dust Mass in Molecular Gas",
                "SmallGrainMassesMolecular": "molecular_small_grain",
                "LargeGrainMassesMolecular": "molecular_large_grain",
                "GraphiteMassesColdDense": "Graphite Dust Mass in Cold-Dense Gas",
                "SilicatesMassesColdDense": "Silicate Dust Mass in Cold-Dense Gas",
                "SmallGrainMassesColdDense": "cold_dense_small_grain",
                "LargeGrainMassesColdDense": "cold_dense_large_grain",
            }[long_species]

        except KeyError:
            raise RegistrationDoesNotMatchError

        full_name = f"{pretty_name} ({aperture_size} kpc)"
        snake_case = f"{short_species}_mass_{aperture_size}_kpc"

        return unit, full_name, snake_case
    else:
        raise RegistrationDoesNotMatchError


def registration_stellar_luminosities(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Registers the luminosities within apertures
    """

    unit = unyt.dimensionless

    # Capture aperture size
    match_string = (
        "Aperture_Luminosities_index_([0-9]?)_aperture_total_star_([0-9]*)_kpc"
    )
    regex = cached_regex(match_string)

    match = regex.match(field_path)

    if match:

        bands = ["u", "g", "r", "i", "z", "Y", "J", "H", "K"]
        band = bands[int(match.group(1))]
        aperture_size = match.group(2)

        full_name = f"{band}-Band Luminosity ({aperture_size} kpc)"
        snake_case = f"{band}_luminosity_{aperture_size}_kpc"

        return unit, full_name, snake_case
    else:
        raise RegistrationDoesNotMatchError


def registration_hydrogen_phase_fractions(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Registers the phase fractions for hydrogen.
    """

    if field_path[:8] != "Hydrogen":
        raise RegistrationDoesNotMatchError

    unit = unyt.dimensionless

    # Need to do a regex search
    # Capture group 1: ionized
    # Capture group 2: massweighted
    # Capture group 3: units
    # Capture group 4: particle type
    match_string = "Hydrogen([a-zA-Z]+)Fractions_([a-zA-Z]+)_([a-zA-Z]+)_?([a-z]*)"
    regex = cached_regex(match_string)
    match = regex.match(field_path)

    if match:
        ionized = match.group(1)
        mass_weighted = match.group(2)
        extracted_units = match.group(3)
        ptype = match.group(4)

        full_name = f"Hydrogen {ionized} Fraction"
        snake_case = ionized.lower()

        return unit, full_name, snake_case
    else:
        raise RegistrationDoesNotMatchError


def registration_black_hole_masses(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Sub-grid black hole property registrations.
    """

    if not field_path[:13] == "SubgridMasses" and field_path[-2:] == "bh":
        raise RegistrationDoesNotMatchError

    unit = unit_system.mass

    # Need to do a regex search
    # Capture group 1: average, min, max.
    # Capture group 2: optional _solar_mass part - backwards compat.
    match_string = "SubgridMasses_([a-z]+)(_solar_mass|)_bh"
    regex = cached_regex(match_string)
    match = regex.match(field_path)

    if match:
        minmax = match.group(1)

        full_name = f"Subgrid Black Hole Mass ({minmax})"
        snake_case = minmax.lower()

        return unit, full_name, snake_case
    else:
        raise RegistrationDoesNotMatchError

    return


def registration_stellar_birth_densities(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Stellar birth density registrations.
    """

    if not field_path[:14] == "BirthDensities" and field_path[-4:] == "star":
        raise RegistrationDoesNotMatchError

    unit = unit_system.mass / unit_system.length ** 3

    # Need to do a regex search (average, min, max)
    match_string = "BirthDensities_([a-z]+)_star"
    regex = cached_regex(match_string)
    match = regex.match(field_path)

    if match:
        minmax = match.group(1)

        full_name = f"Stellar Birth Density ({minmax})"
        snake_case = minmax.lower()

        return unit, full_name, snake_case
    else:
        raise RegistrationDoesNotMatchError

    return


def registration_snii_thermal_feedback_densities(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    SNII thermal feedback density registrations.
    """

    if (
        not field_path[:14] == "DensitiesAtLastSupernovaEvent"
        and field_path[-4:] == "gas"
    ):
        raise RegistrationDoesNotMatchError

    unit = unit_system.mass / unit_system.length ** 3

    # Need to do a regex search (average, min, max)
    match_string = "DensitiesAtLastSupernovaEvent_([a-z]+)_gas"
    regex = cached_regex(match_string)
    match = regex.match(field_path)

    if match:
        minmax = match.group(1)

        full_name = f"SNII Thermal Feedback Density ({minmax})"
        snake_case = minmax.lower()

        return unit, full_name, snake_case
    else:
        raise RegistrationDoesNotMatchError

    return


def registration_species_fractions(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Registers the species mass fraction properties.

    Hopefully this is changed in the future as this is a mess.
    """

    if not field_path[:16] == "SpeciesFractions":
        raise RegistrationDoesNotMatchError

    unit = unyt.dimensionless

    # Need to do a regex search
    # Capture group 1,2: index number - if not present default to 0
    # Capture group 3: mass weighted?
    # Capture group 4: units
    # Capture group 5: particle type
    match_string = (
        "SpeciesFractions(_index_)?([0-9]*)_([a-zA-Z]+)_([a-zA-Z]+)_?([a-z]*)"
    )
    regex = cached_regex(match_string)
    match = regex.match(field_path)

    snake_case = "species"

    if match:
        index = match.group(2) if match.group(2) else 0
        mass_weighted = match.group(3)
        extracted_units = match.group(4)
        ptype = match.group(5)

        full_name = f"Species {index} Fraction"
        snake_case = f"{snake_case}_{index}"

        if ptype:
            cap_ptype = ptype[0].upper() + ptype[1:]
            full_name = f"{cap_ptype} {full_name}"

            snake_case += f"_{ptype}"

    else:
        raise RegistrationDoesNotMatchError

    return unit, full_name, snake_case


def registration_spherical_overdensities(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Registers SO aperture values by searching them with regex.
    """

    # Capture group 1: quantity
    # Capture group 2: particle type
    # Capture group 3: sf / nsf
    # Capture group 4: size of aperture

    match_string = "SO_([^_]*)_([a-zA-Z]*)?_?([a-zA-Z]*)?_?([0-9]*)_rhocrit"
    regex = cached_regex(match_string)

    match = regex.match(field_path)

    if match:
        quantity = match.group(1)
        ptype = match.group(2)
        star_forming = match.group(3)
        aperture_size = int(match.group(4))

        unit = get_aperture_unit(quantity, unit_system)
        name = get_particle_property_name_conversion(quantity, ptype)

        if star_forming:
            sf_in_name = f"{star_forming.upper()} "
        else:
            sf_in_name = ""

        full_name = f"{sf_in_name}{name} ({aperture_size} $\\rho_{{\\rm crit}}$)"
        snake_case = field_path.lower().replace("so_", "")

        return unit, full_name, snake_case
    else:
        raise RegistrationDoesNotMatchError


def registration_element_masses_in_stars(
    field_path: str, unit_system: VelociraptorUnits
) -> (unyt.Unit, str, str):
    """
    Registers element masses contained in stars
    """

    unit = unit_system.mass

    # Capture aperture size
    match_string = "Aperture_([a-zA-Z]*)Masses_aperture_total_star_([0-9]*)_kpc"
    regex = cached_regex(match_string)
    match = regex.match(field_path)

    if match:
        element = match.group(1)
        aperture_size = match.group(2)

        try:
            field = {
                "Oxygen": "Total Oxygen mass in stars",
                "Magnesium": "Total Magnesium mass in stars",
                "Iron": "Total Iron mass in stars",
            }[element]
        except KeyError:
            raise RegistrationDoesNotMatchError

        full_name = f"{field} computed in apertures of size ({aperture_size} kpc)"

        snake_case = f"{element.lower()}_mass_{aperture_size}_kpc"

        return unit, full_name, snake_case
    else:
        raise RegistrationDoesNotMatchError

    return


# TODO
# lambda_B
# q
# q_gas
# q_star
# s
# s_gas
# s_star
# sigV
# sigV_gas_nsf
# sigV_gas_sf


# This must be placed at the bottom of the file so that we
# have defined all functions before getting to it.
# This dictionary will be turned into sets of datasets that
# contain the results of the registraiton functions. For example.
# we will have VelociraptorProperties.energies.erot for the rotation
# energy.
global_registration_functions = {
    k: globals()[f"registration_{k}"]
    for k in [
        "metallicity",
        "ids",
        "energies",
        "stellar_age",
        "spherical_overdensities",
        "rotational_support",
        "star_formation_rate",
        "masses",
        "eigenvectors",
        "radii",
        "temperature",
        "veldisp",
        "structure_type",
        "velocities",
        "positions",
        "concentration",
        "rvmax_quantities",
        "angular_momentum",
        "projected_apertures",
        "apertures",
        "element_mass_fractions",
        "dust_mass_fractions",
        "number",
        "hydrogen_phase_fractions",
        "black_hole_masses",
        "stellar_birth_densities",
        "snii_thermal_feedback_densities",
        "species_fractions",
        "gas_hydrogen_species_masses",
        "gas_H_and_He_masses",
        "gas_diffuse_element_masses",
        "dust_masses_from_table",
        "dust_masses",
        "stellar_luminosities",
        "cold_dense_gas_properties",
        "log_element_ratios_times_masses",
        "lin_element_ratios_times_masses",
        "element_masses_in_stars",
        "fail_all",
    ]
}



"""
Sub-module for adding observational data to plots.

Includes the ObservationalData object and helper functions
to convert data to this new format.
"""

from velociraptor.observations.objects import (
    ObservationalData,
    MultiRedshiftObservationalData,
)
from velociraptor.exceptions import ObservationalDataError

from typing import Union, List, Iterable
from warnings import warn


def load_observation(filename: str):
    """
    Load an observation from file filename. This should be in the
    standard velociraptor format.

    Deprecated in favour of :func:`load_observations`

    Parameters
    ----------

    filename: str
        Filename of the observational dataset that you wish to load.
        Should probably end in .hdf5. See the documentation for
        :class:`velociraptor.observations.objects.ObservationalData`
        for more information.

    Returns
    -------

    velociraptor.observations.objects.ObservationalData:
        Observational data instance read from file.
    """

    warn(
        "load_observation is deprecated and will be removed in a future version. "
        "Please use load_observations that can load data from multi-redshift "
        "datasets introduced in velociraptor v0.12.0.",
        DeprecationWarning,
    )

    data = ObservationalData()
    data.load(filename)

    return data


def load_observations(
    filenames: Union[str, Iterable[str]], redshift_bracket: List[float] = [0.0, 1000.0]
):
    """
    Parameters
    ----------

    filename: str, Iterable[str]
        Filename(s) of the observational dataset that you wish to load.
        Should probably end in .hdf5. See the documentation for
        :class:`velociraptor.observations.objects.ObservationalData`
        and :class:`velociraptor.observations.objects.MultiRedshiftObservationalData`
        for more information.

    redshift_bracket: str
        Redshift bracket to overlap with. If any of the observations in the
        file overlap with this bracket, they are returned. By default, this
        bracket is 0.0 to 1000.0, so will encompass all reasonable
        observations present in the file.
    

    Returns
    -------

    List[velociraptor.observations.objects.ObservationalData]:
        Observational data instances read from file that overlap with your
        specified redshift bracket.
    """

    returned_data = []

    if not isinstance(filenames, list):
        filenames = [filenames]

    for filename in filenames:
        try:
            multi_z = MultiRedshiftObservationalData()
            multi_z.load(filename)

            returned_data += multi_z.get_datasets_overlapping_with(
                redshifts=redshift_bracket
            )
        except ObservationalDataError:
            data = ObservationalData()
            data.load(filename)

            # Lower and upper bounds of the requested redshift bracket to
            # return datasets between
            lower, upper = redshift_bracket

            if (
                (data.redshift_lower <= lower and lower <= data.redshift_upper)
                or (data.redshift_lower <= upper and upper <= data.redshift_upper)
                or (lower <= data.redshift_lower and data.redshift_upper <= upper)
            ):
                returned_data.append(data)

    return returned_data



"""
Caching functions for regexes.
"""

from functools import lru_cache
import re


@lru_cache(64)
def cached_regex(regex_match_string: str, flags: int = 0):
    return re.compile(regex_match_string, flags)



"""
SWIFTsimIO integration.

This allows users to load the contents of various haloes as `swiftsimio`
datasets in a computationally efficient way.
"""


import swiftsimio
import numpy as np

from velociraptor.particles.particles import VelociraptorParticles

from typing import Union, Tuple

from collections import namedtuple


def generate_spatial_mask(
    particles: VelociraptorParticles, snapshot_filename
) -> swiftsimio.mask:
    """
    Determines the mask defining the spatial region containing
    the particles of interest.

    This uses `r_max` to define the extent of the region, so you
    will need to instantiate the VelociraptorParticles instance
    with an associated catalogue to use this feature, as it requires
    the knowledge of `r_max`.

    Takes two arguments:

    + particles, the VelociraptorParticles instance,
    + snapshot_filename, the path to the associated SWIFT snapshot.

    It returns:

    + mask, an object containing masks for all available datasets in the
            swift dataset.
    """

    # First use the swiftsimio spatial masking to constrain our dataset
    # to only contain particles within the cube that contains the halo
    # (this is only approximate down to the swift cell size)
    swift_mask = swiftsimio.mask(snapshot_filename, spatial_only=True)

    # SWIFT data is stored in comoving units, so we need to un-correct
    # the velociraptor data if it is stored in physical.
    try:
        if not particles.groups_instance.catalogue.units.comoving:
            length_factor = particles.groups_instance.catalogue.units.a
        else:
            length_factor = 1.0
    except AttributeError:
        raise RuntimeError(
            "Please use a particles instance with an associated halo " "catalogue."
        )

    spatial_mask = [
        [
            particles.x / length_factor - particles.r_size / length_factor,
            particles.x / length_factor + particles.r_size / length_factor,
        ],
        [
            particles.y / length_factor - particles.r_size / length_factor,
            particles.y / length_factor + particles.r_size / length_factor,
        ],
        [
            particles.z / length_factor - particles.r_size / length_factor,
            particles.z / length_factor + particles.r_size / length_factor,
        ],
    ]

    swift_mask.constrain_spatial(spatial_mask)

    return swift_mask


def generate_bound_mask(
    data: swiftsimio.reader.SWIFTDataset, particles: VelociraptorParticles
) -> namedtuple:
    """
    Determines the mask defining the particles bound to the object
    of interest.

    Takes two arguments:

    + data, a SWIFTDataset, which may be spatially masked,
    + particles, the VelociraptorParticles instance.

    It returns:

    + mask, an object containing masks for all available datasets in the
            swift dataset.
    """
    # Now we must generate the secondary mask, for all available
    # particle types.

    particle_name_masks = {}

    for particle_name in data.metadata.present_particle_names:
        # This will change if we ever take advantage of the
        # parttypes available through velociraptor.
        particle_name_masks[particle_name] = np.in1d(
            getattr(data, particle_name).particle_ids, particles.particle_ids
        )

    # Finally we generate a named tuple with the correct fields and
    # fill it with the contents of our dictionary
    MaskTuple = namedtuple("MaskCollection", data.metadata.present_particle_names)
    mask = MaskTuple(**particle_name_masks)

    return mask


def to_swiftsimio_dataset(
    particles: VelociraptorParticles,
    snapshot_filename,
    generate_extra_mask: bool = False,
) -> Union[
    swiftsimio.reader.SWIFTDataset, Tuple[swiftsimio.reader.SWIFTDataset, namedtuple]
]:
    """
    Loads a VelociraptorParticles instance for one halo into a
    `swiftsimio` masked dataset.

    Initially, this uses `r_max` to perform a spatial mask, and
    then returns the `swiftsimio` dataset and a secondary mask
    that may be used to extract only the particles that are
    part of the FoF group.

    You will need to instantiate the VelociraptorParticles instance
    with an associated catalogue to use this feature, as it requires
    the knowledge of `r_max`.

    Takes three arguments:

    + particles, the VelociraptorParticles instance,
    + snapshot_filename, the path to the associated SWIFT snapshot.
    + generate_extra_mask, whether or not to generate the secondary
                           mask object that allows for the extraction
                           of particles that are present only in the
                           FoF group.

    It returns:

    + data, the swiftsimio dataset
    + mask, an object containing masks for all available datasets in the
            swift dataset. The initial masking is performed on a
            spatial only basis, and this is required to only extract
            the particles in the FoF group as identified by
            velociraptor. This is only provided if generate_extra_mask
            has a truthy value.
    """

    swift_mask = generate_spatial_mask(particles, snapshot_filename)

    data = swiftsimio.load(snapshot_filename, mask=swift_mask)

    if not generate_extra_mask:
        return data

    mask = generate_bound_mask(data, particles)

    return data, mask



"""
Tools sub-module for generating information for plotting.
"""

from velociraptor.tools.labels import get_full_label, get_mass_function_label
from velociraptor.tools.lines import binned_mean_line, binned_median_line
from velociraptor.tools.mass_functions import create_mass_function



"""
Tools for creating mass functions!
"""

import unyt
import numpy as np

from velociraptor.tools.labels import get_mass_function_label_no_units


def create_mass_function_given_bins(
    masses: unyt.unyt_array,
    bins: unyt.unyt_array,
    box_volume: unyt.unyt_quantity,
    minimum_in_bin: int = 3,
):
    """
    Creates a mass function (with equal width bins in log M) for you to plot.

    Parameters
    ----------

    masses: unyt.unyt_array
        The array that you want to create a mass function of (usually this is
        for example halo masses or stellar masses).

    bins: unyt.unyt_array
        The mass bin edges to use.

    unyt.unyt_quantity: box_volume
        The volume of the box such that we can return ``n / volume``.

    minimum_in_bin: int, optional
        The number of objects in a bin for it to be classed as valid. Bins
        with a number of objects smaller than this are not returned. By default
        this parameter takes a value of 3.


    Returns
    -------

    bin_centers: unyt.unyt_array
        The centers of the bins (taken to be the linear mean of the bin edges).

    mass_function: unyt.unyt_array
        The value of the mass function at the bin centers.

    error: unyt.unyt_array
        Scatter in the mass function (Poisson errors).

    """

    bins.convert_to_units(masses.units)

    # This is required to ensure that the mass function converges with bin width
    bin_width_in_logspace = np.log10(bins[1]) - np.log10(bins[0])
    normalization_factor = 1.0 / (bin_width_in_logspace * box_volume)

    mass_function, _ = np.histogram(masses, bins)
    valid_bins = mass_function >= minimum_in_bin

    # Poisson sampling
    error = np.sqrt(mass_function)

    mass_function *= normalization_factor
    error *= normalization_factor

    bin_centers = 0.5 * (bins[1:] + bins[:-1])

    mass_function.name = get_mass_function_label_no_units("{}")
    bin_centers.name = masses.name

    return bin_centers[valid_bins], mass_function[valid_bins], error[valid_bins]


def create_mass_function(
    masses: unyt.unyt_array,
    lowest_mass: unyt.unyt_quantity,
    highest_mass: unyt.unyt_quantity,
    box_volume: unyt.unyt_quantity,
    n_bins: int = 25,
    minimum_in_bin: int = 3,
    return_bin_edges: bool = False,
):
    """
    Creates a mass function (with equal width bins in log M) for you to plot.

    Parameters
    ----------

    masses: unyt.unyt_array
        The array that you want to create a mass function of (usually this is
        for example halo masses or stellar masses).

    lowest_mass: unyt.unyt_quantity
        the lowest mass edge of the bins
    
    highest_mass: unyt.unyt_quantity
        the highest mass edge of the bins

    box_volume: unyt.unyt_quantity
        The volume of the box such that we can return ``n / volume``.

    n_bins: unyt.unyt_array
        The number of equal log-width bins across the range to use.

    minimum_in_bin: int, optional
        The number of objects in a bin for it to be classed as valid. Bins
        with a number of objects smaller than this are not returned. By default
        this parameter takes a value of 3.

    return_bin_edges: bool, optional
        Return the bin edges used in the binning process? Default is False.

    Returns
    -------

    bin_centers: unyt.unyt_array
        The centers of the bins (taken to be the linear mean of the bin edges).

    mass_function: unyt.unyt_array
        The value of the mass function at the bin centers.

    error: unyt.unyt_array
        Scatter in the mass function (Poisson errors).

    bin_edges: unyt.unyt_array, optional
        Bin edges that were used in the binning process.

    """

    assert (
        masses.units == lowest_mass.units and lowest_mass.units == highest_mass.units
    ), "Please ensure that all mass quantities have the same units."

    bins = (
        np.logspace(np.log10(lowest_mass), np.log10(highest_mass), n_bins + 1)
        * masses.units
    )

    bin_centers, mass_function, error = create_mass_function_given_bins(
        masses=masses, bins=bins, box_volume=box_volume, minimum_in_bin=minimum_in_bin
    )

    if return_bin_edges:
        return bin_centers, mass_function, error, bins
    else:
        return bin_centers, mass_function, error


def create_adaptive_mass_function(
    masses: unyt.unyt_array,
    lowest_mass: unyt.unyt_quantity,
    highest_mass: unyt.unyt_quantity,
    box_volume: unyt.unyt_quantity,
    base_n_bins: int = 25,
    minimum_in_bin: int = 3,
    return_bin_edges: bool = False,
):
    """
    Creates a mass function (with equal width bins in log M) for you to plot.

    Parameters
    ----------

    masses: unyt.unyt_array
        The array that you want to create a mass function of (usually this is
        for example halo masses or stellar masses).

    lowest_mass: unyt.unyt_quantity
        the lowest mass edge of the bins
    
    highest_mass: unyt.unyt_quantity
        the highest mass edge of the bins

    box_volume: unyt.unyt_quantity
        The volume of the box such that we can return ``n / volume``.

    base_n_bins: unyt.unyt_array
        The number of equal log-width bins across the range to use in the case
        where no adaptive sampling is required. This returns the minimal allowed
        bin width.

    minimum_in_bin: int, optional
        The number of objects in a bin for it to be classed as valid. Bins
        with a number of objects smaller than this are not returned. By default
        this parameter takes a value of 3.

    return_bin_edges: bool, optional
        Return the bin edges used in the binning process? Default is False.

    Returns
    -------

    bin_centers: unyt.unyt_array
        The centers of the bins (taken to be the linear mean of the bin edges).

    mass_function: unyt.unyt_array
        The value of the mass function at the bin centers.

    error: unyt.unyt_array
        Scatter in the mass function (Poisson errors).

    bin_edges: unyt.unyt_array, optional
        Bin edges that were used in the binning process.

    """
    assert (
        masses.units == lowest_mass.units and lowest_mass.units == highest_mass.units
    ), "Please ensure that all mass quantities have the same units."

    # We must do this and not modify the data internally as it could be used for
    # plotting elsewhere.
    sorted_masses = np.sort(masses)

    mask = np.logical_and(sorted_masses >= lowest_mass, sorted_masses <= highest_mass)

    # Also, we bin in logspace.
    sorted_masses = np.log10(sorted_masses[mask])

    minimal_log_bin_width = (
        np.log10(highest_mass) - np.log10(lowest_mass)
    ) / base_n_bins

    number_in_bin = []
    bin_edges_left = [np.log10(lowest_mass)]
    bin_edges_right = []
    bin_medians = []

    current_edge_left = bin_edges_left[0]
    current_lower_index = 0
    current_bin_count = 0

    for index, mass in enumerate(sorted_masses):
        current_bin_count += 1

        if mass < current_edge_left + minimal_log_bin_width:
            # Nothing to do, we're just filling the bin
            continue
        elif current_bin_count > minimum_in_bin:
            # Let's just end this here
            bin_medians.append(
                np.median(sorted_masses[current_lower_index : index + 1])
            )

            # The new bin edge lives half way in between our current value and the
            # next value, if it exists.
            try:
                new_edge = 0.5 * (sorted_masses[index + 1] + mass)
            except IndexError:
                # This case is where `value` is the last item in the array.
                new_edge = mass

            bin_edges_right.append(new_edge)
            bin_edges_left.append(new_edge)
            number_in_bin.append(current_bin_count)

            # Reset for the lads
            current_edge_left = new_edge
            current_lower_index = index
            current_bin_count = 0
        else:
            # Nothing we can do! Let's just keep going and
            # hope we find more matches.
            pass

    # Clean up the last bin? We're gonna miss the largest N galaxies..
    if current_bin_count > 1:
        # We can create a novel bin!
        bin_edges_right.append(mass)
        bin_medians.append(np.median(sorted_masses[current_lower_index : index + 1]))
        number_in_bin.append(current_bin_count)
    elif current_bin_count == 1:
        # Extend the previous bin (otherwise error is consistent with zero)
        try:
            bin_edges_right[-1] = mass
            number_in_bin[-1] += 1
            bin_medians[-1] = np.median(sorted_masses[-number_in_bin[-1] :])
            # We don't need the next bin now.
            bin_edges_left = bin_edges_left[:-1]

        # There is no previous bin because we have just one bin in total
        except IndexError:
            bin_edges_right.append(mass)
            number_in_bin.append(1)
            bin_medians.append(sorted_masses[-1])
    else:
        # This bin doesn't exist anyway, boo...
        bin_edges_left = bin_edges_left[:-1]

    bin_widths = [right - left for right, left in zip(bin_edges_right, bin_edges_left)]

    mass_function = [
        n / (width * box_volume) for n, width in zip(number_in_bin, bin_widths)
    ]
    error = [
        np.sqrt(n) / (width * box_volume) for n, width in zip(number_in_bin, bin_widths)
    ]

    try:
        mass_function_units = mass_function[0].units
    except IndexError:
        mass_function_units = (1 / box_volume).units

    try:
        error_units = error[0].units
    except IndexError:
        error_units = mass_function_units

    mass_function = unyt.unyt_array(mass_function, units=mass_function_units)
    # For some reason when using a name= argument here this ends up as None?
    mass_function.name = get_mass_function_label_no_units("{}")

    error = unyt.unyt_array(error, units=error_units)

    bin_centers = unyt.unyt_array(
        [10 ** x for x in bin_medians], units=masses.units, name=masses.name
    )

    if return_bin_edges:
        return (
            bin_centers,
            mass_function,
            error,
            unyt.unyt_array(
                10 ** np.array([bin_edges_left, bin_edges_right]), units=masses.units
            ),
        )
    else:
        return bin_centers, mass_function, error



"""
Fitting formulae for the stellar mass-halo mass relation.
"""

"""
Creates a stellar mass function plot for various velociraptor
catalogues.
"""

import unyt
import numpy as np

from velociraptor.catalogue.catalogue import VelociraptorCatalogue

from typing import Union, List


def moster(
    catalogue: VelociraptorCatalogue,
    mass_range: Union[unyt.unyt_array, List[unyt.unyt_quantity]] = [
        1e4 * unyt.msun,
        1e16 * unyt.msun,
    ],
    n_eval: int = 256,
):
    """
    The moster retaion (from Moster+ 2013). Original code provided by
    Matthieu Schaller. Takes:

    + catalogue, your velociraptor catalogue object
    + mass_range, a length 2 array with the lowest and highest halo mass
                  you would like the relation to be evaluated between
                  (default: 1e4 msun to 1e16 msun)
    + n_eval, the number of function evaluations (equally spaced in log (Mhalo))
              (default: 256)

    Returns:

    + halo_mass, the halo masses at which the relation is evaluated at
    + stellar_mass, the stellar masses matching with the above halo masses
    """

    z = catalogue.z
    halo_mass = (
        np.logspace(
            np.log10(mass_range[0].value), np.log10(mass_range[1].value), n_eval
        )
        * mass_range[0].units
    )

    stellar_mass = moster_raw(z, halo_mass.value) * halo_mass.units

    return halo_mass, stellar_mass


def moster_raw(z, Mhalo):
    """
    Stellar mass-halo mass relation from Moster+2013.

    Provided by Matthieu Schaller.
    """

    # Table 1
    M_10 = 11.590
    M_11 = 1.195
    N_10 = 0.0351
    N_11 = -0.0247
    beta_10 = 1.376
    beta_11 = -0.826
    gama_10 = 0.608
    gama_11 = 0.329

    # Equations 11 - 14
    log_M_1 = M_10 + M_11 * z / (z + 1)
    N = N_10 + N_11 * z / (z + 1)
    beta = beta_10 + beta_11 * z / (z + 1)
    gama = gama_10 + gama_11 * z / (z + 1)

    M_1 = 10 ** log_M_1

    # Equation 2
    Mstar_over_Mhalo = 2.0 * N * ((Mhalo / M_1) ** -beta + (Mhalo / M_1) ** gama) ** -1

    return Mstar_over_Mhalo * Mhalo


def behroozi(
    catalogue: VelociraptorCatalogue,
    mass_range: Union[unyt.unyt_array, List[unyt.unyt_quantity]] = [
        1e4 * unyt.msun,
        1e16 * unyt.msun,
    ],
    n_eval: int = 256,
):
    """
    The behroozi fit to the SMHMR (from Behroozi+ 2013). Original code provided by
    Matthieu Schaller. Takes:

    + catalogue, your velociraptor catalogue object
    + mass_range, a length 2 array with the lowest and highest halo mass
                  you would like the relation to be evaluated between
                  (default: 1e4 msun to 1e16 msun)
    + n_eval, the number of function evaluations (equally spaced in log (Mhalo))
              (default: 256)

    Returns:

    + halo_mass, the halo masses at which the relation is evaluated at
    + stellar_mass, the stellar masses matching with the above halo masses
    """

    z = catalogue.z
    halo_mass = (
        np.logspace(
            np.log10(mass_range[0].value), np.log10(mass_range[1].value), n_eval
        )
        * mass_range[0].units
    )

    stellar_mass = behroozi_raw(z, halo_mass.value) * halo_mass.units

    return halo_mass, stellar_mass


def behroozi_raw(z, Mhalo):
    """
    Stellar mass-halo mass relation from Behroozi +2013.

    Provided by Matthieu Schaller.
    """
    aexp = 1.0 / (1.0 + z)
    nu = np.exp(-4.0 * aexp ** 2)
    mh = Mhalo
    m1_0 = 11.514
    m1_a = -1.793
    m1_z = -0.251
    m1 = 10.0 ** (m1_0 + nu * (m1_a * (aexp - 1.0) + m1_z * z))
    eps0 = -1.777
    eps_a = -0.006
    eps_z = 0.000
    eps_a2 = -0.119
    eps = 10.0 ** (
        eps0 + nu * (eps_a * (aexp - 1.0) + eps_z * z) + eps_a2 * (aexp - 1.0)
    )
    alpha_0 = -1.412
    alpha_a = 0.731
    alpha = alpha_0 + nu * (alpha_a * (aexp - 1.0))
    delta_0 = 3.508
    delta_a = 2.608
    delta_z = -0.043
    delta = delta_0 + nu * (delta_a * (aexp - 1.0) + delta_z * z)
    gamma_0 = 0.316
    gamma_a = 1.319
    gamma_z = 0.279
    gamma = gamma_0 + nu * (gamma_a * (aexp - 1.0) + gamma_z * z)
    x = np.log10(mh / m1)
    f = -1.0 * np.log10(10.0 ** (alpha * x) + 1.0) + delta * (
        np.log10(1.0 + np.exp(x))
    ) ** gamma / (1.0 + np.exp(10.0 ** (-1.0 * x)))
    x0 = 0.0
    f0 = -1.0 * np.log10(10.0 ** (alpha * x0) + 1.0) + delta * (
        np.log10(1.0 + np.exp(x0))
    ) ** gamma / (1.0 + np.exp(10.0 ** (-1.0 * x0)))
    logmstar = np.log10(eps * m1) + f - f0
    return 10.0 ** logmstar


def behroozi_2019_raw(z, Mhalo):
    """
    Stellar mass-halo mass relation from Behroozi +2019.

    The data is taken from https://www.peterbehroozi.com/data.html

    This function is a median fit to the raw data for centrals (i.e. excluding
    satellites)

    The stellar mass is the true stellar mass (i.e. w/o observational corrections)

    The fitting function does not include the intrahalo light contribution to the
    stellar mass

    The halo mass is the peak halo mass that follows the Bryan & Norman (1998)
    spherical overdensity definition

    Provided by Evgenii Chaikin.
    """

    Mhalo_log = np.log10(Mhalo)

    params = {
        "EFF_0": -1.431495,
        "EFF_0_A": 1.75703,
        "EFF_0_A2": 1.350451,
        "EFF_0_Z": -0.217846,
        "M_1": 12.07402,
        "M_1_A": 4.599896,
        "M_1_A2": 4.423389,
        "M_1_Z": -0.7324986,
        "ALPHA": 1.973839,
        "ALPHA_A": -2.468417,
        "ALPHA_A2": -1.816299,
        "ALPHA_Z": 0.18208,
        "BETA": 0.4702271,
        "BETA_A": -0.8751643,
        "BETA_Z": -0.486642,
        "DELTA": 0.3822958,
        "GAMMA": -1.160189,
        "GAMMA_A": -3.633671,
        "GAMMA_Z": -1.2189,
    }

    a = 1.0 / (1.0 + z)
    a1 = a - 1.0
    lna = np.log(a)

    zparams = {}

    zparams["m_1"] = (
        params["M_1"]
        + a1 * params["M_1_A"]
        - lna * params["M_1_A2"]
        + z * params["M_1_Z"]
    )
    zparams["sm_0"] = (
        zparams["m_1"]
        + params["EFF_0"]
        + a1 * params["EFF_0_A"]
        - lna * params["EFF_0_A2"]
        + z * params["EFF_0_Z"]
    )
    zparams["alpha"] = (
        params["ALPHA"]
        + a1 * params["ALPHA_A"]
        - lna * params["ALPHA_A2"]
        + z * params["ALPHA_Z"]
    )
    zparams["beta"] = params["BETA"] + a1 * params["BETA_A"] + z * params["BETA_Z"]
    zparams["delta"] = params["DELTA"]
    zparams["gamma"] = 10 ** (
        params["GAMMA"] + a1 * params["GAMMA_A"] + z * params["GAMMA_Z"]
    )

    dm = Mhalo_log - zparams["m_1"]
    dm2 = dm / zparams["delta"]
    logmstar = (
        zparams["sm_0"]
        - np.log10(10 ** (-zparams["alpha"] * dm) + 10 ** (-zparams["beta"] * dm))
        + zparams["gamma"] * np.exp(-0.5 * (dm2 * dm2))
    )

    return 10.0 ** logmstar



"""
Objects for the particles files. This includes:

+ catalog_groups
+ catalog_particles
+ catalog_parttypes

and their unbound variants.

The objects defined here are as follows:

VelociraptorGroups: This depends on the VelociraptorCatalogue object
                    and allows for users to find information about where
                    in each file each group belongs.
VelociraptorParticles: This depends on VelociraptorGroups and itself allows
                       for the particles in a _single group_ to be loaded.
                       Along with this, we also load the catalog_parttypes
                       file for those selected particles.
"""

import h5py
import numpy as np


from velociraptor.catalogue.catalogue import VelociraptorCatalogue

from typing import Union, Dict


class VelociraptorGroups(object):
    """
    Groups object for velociraptor. Contains the information about
    where the groups start and stop in the catalog_particles file.
    """

    # Forward declarations of numbers

    # Attributes
    file_id: np.int64
    num_of_files: np.int64
    num_of_groups: np.int64
    total_num_of_groups: np.int64

    # Arrays read from HDF5
    group_size: np.ndarray
    number_of_substructures_in_halo: np.ndarray
    offset: np.ndarray
    offset_unbound: np.ndarray
    parent_halo_id: np.ndarray

    def __init__(self, filename, catalogue: Union[VelociraptorCatalogue, None]):
        """
        Initialise the velociraptor groups object. This
        loads all of the groups data, that should be relatively
        small, into several properties. We then provide functions
        to extract VelociraptorParticles objects for each individual
        halo.
        """

        if catalogue.mask is not Ellipsis:
            raise ValueError("VelociraptorGroups not compatible with masked catalogue.")

        self.filename = filename
        self.catalogue = catalogue

        self.__load_attributes()

        return

    def __load_attributes(self):
        """
        Loads the attributes from the HDF5 file.
        """

        read_to_attribute = [
            "File_id",
            "Num_of_files",
            "Num_of_groups",
            "Total_num_of_groups",
        ]

        read_to_array = [
            "Group_Size",
            "Number_of_substructures_in_halo",
            "Offset",
            "Offset_unbound",
            "Parent_halo_ID",
        ]

        with h5py.File(self.filename, "r") as handle:
            for attribute in read_to_attribute:
                setattr(self, attribute.lower(), handle[attribute][0])

            for attribute in read_to_array:
                try:
                    setattr(self, attribute.lower(), handle[attribute][:])
                except KeyError:
                    # Whoops! Not available.
                    pass

        return

    def extract_halo(
        self, halo_index: int, filenames: Union[Dict[str, str], None] = None
    ):
        """
        Get a halo particles object for a given index into the catalogue (NOT the halo unique
        id). Filenames is either a dictionary with the following structure:

        {
            "particles_filename": "...",
            "parttypes_filename": "...",
            "unbound_particles_filename": "...",
            "unbound_parttypes_filename": "...",
        }

        or None, in which case we guess what the filename should
        be from the filename of the groups that has already been passed.
        """

        if filenames is None:
            particles_filename = str(self.filename).replace(
                "catalog_groups", "catalog_particles"
            )
            parttypes_filename = str(self.filename).replace(
                "catalog_groups", "catalog_parttypes"
            )
            unbound_particles_filename = str(self.filename).replace(
                "catalog_groups", "catalog_particles.unbound"
            )
            unbound_parttypes_filename = str(self.filename).replace(
                "catalog_groups", "catalog_parttypes.unbound"
            )
        else:
            particles_filename = filenames["particles_filename"]
            parttypes_filename = filenames["parttypes_filename"]
            unbound_particles_filename = filenames["unbound_particles_filename"]
            unbound_parttypes_filename = filenames["unbound_parttypes_filename"]

        if halo_index == self.offset.size - 1:  # last halo in catalog
            with h5py.File(particles_filename) as particles_file:
                total_particles = particles_file[
                    "Total_num_of_particles_in_all_groups"
                ][0]
                number_of_particles = total_particles - self.offset[halo_index]
            with h5py.File(unbound_particles_filename) as unbound_particles_file:
                total_unbound_particles = unbound_particles_file[
                    "Total_num_of_particles_in_all_groups"
                ][0]
                number_of_unbound_particles = (
                    total_unbound_particles - self.offset_unbound[halo_index]
                )
        else:
            number_of_particles = self.offset[halo_index + 1] - self.offset[halo_index]
            number_of_unbound_particles = (
                self.offset_unbound[halo_index + 1] - self.offset_unbound[halo_index]
            )
        assert (
            number_of_particles + number_of_unbound_particles
            == self.group_size[halo_index]
        ), "Something is incorrect in the calculation of group sizes for halo {}. Group_Size: {}, Bound: {}, Unbound: {}".format(
            halo_index,
            self.group_size[halo_index],
            number_of_particles,
            number_of_unbound_particles,
        )

        particles = VelociraptorParticles(
            particles_filename=particles_filename,
            parttypes_filename=parttypes_filename,
            offset=self.offset[halo_index],
            group_size=number_of_particles,
            groups_instance=self,
        )

        unbound_particles = VelociraptorParticles(
            particles_filename=unbound_particles_filename,
            parttypes_filename=unbound_parttypes_filename,
            offset=self.offset_unbound[halo_index],
            group_size=number_of_unbound_particles,
            groups_instance=self,
        )

        if self.catalogue is not None:
            particles.register_halo_attributes(self.catalogue, halo_index)
            unbound_particles.register_halo_attributes(self.catalogue, halo_index)

        return particles, unbound_particles


class VelociraptorParticles(object):
    """
    Velociraptor particles object, holds information on a single
    halo's particles, including which IDs and particle types are
    in that halo. This provides extra post-processing options, such
    as splitting the IDs by particle type.
    """

    def __init__(
        self,
        particles_filename,
        parttypes_filename,
        offset: int,
        group_size: int,
        groups_instance: VelociraptorGroups,
    ):
        """
        Takes:

        + particles filename, the filename of the .catalog_particles file
        + parttype filename, the filename of the .catalog_parttypes file
        + offset, the offset from the .catalog_groups file
        + group_size, the size of the group in number of particles.
        + groups_instance, the associated groups instance
        """

        self.particles_filename = particles_filename
        self.parttypes_filename = parttypes_filename
        self.offset = offset
        self.group_size = group_size
        self.groups_instance = groups_instance

        self.__load_particles()
        self.__load_parttypes()

        return

    def __load_particles(self):
        """
        Load the information from the .catalog_particles file.
        """

        read_to_attribute = [
            "File_id",
            "Num_of_files",
            "Num_of_particles_in_groups",
            "Total_num_of_particles_in_all_groups",
        ]

        with h5py.File(self.particles_filename, "r") as handle:
            for attribute in read_to_attribute:
                setattr(self, f"particle_{attribute.lower()}", handle[attribute][0])

            # Load only the particle ids that we actually need
            self.particle_ids = handle["Particle_IDs"][
                self.offset : self.offset + self.group_size
            ]

        return

    def __load_parttypes(self):
        """
        Load the information from the .catalog_parttypes file.
        """

        read_to_attribute = [
            "File_id",
            "Num_of_files",
            "Num_of_particles_in_groups",
            "Total_num_of_particles_in_all_groups",
        ]

        with h5py.File(self.parttypes_filename, "r") as handle:
            for attribute in read_to_attribute:
                setattr(self, f"parttypes_{attribute.lower()}", handle[attribute][0])

            # Load only the particle ids that we actually need
            self.particle_types = handle["Particle_types"][
                self.offset : self.offset + self.group_size
            ]

        return

    def register_halo_attributes(
        self, catalogue: VelociraptorCatalogue, halo_index: int
    ):
        """
        Registers useful halo attributes to this object (such as the mass and radii of the halo).
        """

        self.halo_index = halo_index

        self.mass_200crit = catalogue.masses.mass_200crit[halo_index]
        self.mass_200mean = catalogue.masses.mass_200mean[halo_index]
        self.mass_bn98 = catalogue.masses.mass_bn98[halo_index]
        self.mass_fof = catalogue.masses.mass_fof[halo_index]
        self.mvir = catalogue.masses.mvir[halo_index]

        self.r_200crit = catalogue.radii.r_200crit[halo_index]
        self.r_200mean = catalogue.radii.r_200mean[halo_index]
        self.r_bn98 = catalogue.radii.r_bn98[halo_index]
        self.r_size = catalogue.radii.r_size[halo_index]
        self.rmax = catalogue.radii.rmax[halo_index]
        self.rvir = catalogue.radii.rvir[halo_index]

        self.x = catalogue.positions.xc[halo_index]
        self.y = catalogue.positions.yc[halo_index]
        self.z = catalogue.positions.zc[halo_index]

        # x_gas and x_star are measured relative to xc for some reason.
        # The following may not be available in the catalogues.
        try:
            self.x_gas = catalogue.positions.xc_gas[halo_index] + self.x
            self.y_gas = catalogue.positions.yc_gas[halo_index] + self.y
            self.z_gas = catalogue.positions.zc_gas[halo_index] + self.z
        except AttributeError:
            pass

        try:
            self.x_star = catalogue.positions.xc_star[halo_index] + self.x
            self.y_star = catalogue.positions.yc_star[halo_index] + self.y
            self.z_star = catalogue.positions.zc_star[halo_index] + self.z
        except AttributeError:
            pass

        try:
            self.x_mbp = catalogue.positions.xcmbp[halo_index]
            self.y_mbp = catalogue.positions.ycmbp[halo_index]
            self.z_mbp = catalogue.positions.zcmbp[halo_index]
        except AttributeError:
            pass

        return



"""
Tools for creating comparison plots that are built out of the saved data.yml files.
"""

"""
A script for plotting multiple data lines on one set of observational data.
This uses the `data.yml` files from the pipeline.
See the help (from argparse) for more information. That can be seen by using
python3 create_comparison.py -h
"""

import matplotlib.pyplot as plt
import yaml
import unyt

from numpy import log10
from typing import Union, List, Dict, Optional, Tuple
from velociraptor.autoplotter.objects import VelociraptorPlot, valid_line_types
from velociraptor.autoplotter.plot import decorate_axes

from velociraptor.autoplotter.objects import AutoPlotter
from velociraptor.autoplotter.metadata import AutoPlotterMetadata
from velociraptor.observations import load_observations


class FakeCatalogue(object):
    """
    Fake VelociraptorCatalogue used to store redshift and
    scale factor information if available.
    """

    def __init__(self, z=0.0, a=0.0):
        self.z = float(z)
        self.a = float(a)
        return


def load_yaml_line_data(
    paths: Union[str, List[str]], names: Union[str, List[str]]
) -> Dict[str, Dict]:
    """
    Load the ``yaml`` data from the files for lines.
    Parameters
    ----------
    paths: Union[str, List[str]]
        Paths to yaml data files to load.

    names: Union[str, List[str]]
        Names of the simulations that correspond to the yaml data files.
        Will be placed in the legends of the plots.

    Returns
    -------

    data: Dict[str, Dict]
        Dictionary of line data read directly from the files.
    """

    if not isinstance(paths, list):
        paths = [paths]
        names = [names]

    data = {}

    for path, name in zip(paths, names):
        try:
            with open(path, "r") as handle:
                data[name] = yaml.load(handle, Loader=yaml.Loader)
        except (OSError, FileNotFoundError):
            data[name] = {}

    return data


def recreate_instances(
    config: Union[str, List[str]],
    paths: Union[str, List[str]],
    names: Union[str, List[str]],
    observational_data_directory: Optional[str] = None,
    file_extension: Optional[str] = None,
) -> Tuple[AutoPlotter, Dict[str, Dict]]:
    """
    Recreates instances of required objects for passing to
    ``reacreate_single_figure``.

    Parameters
    ----------

    config: str, List[str],
        The autoplotter config files that are associated with the yaml data below.

    paths: str, List[str],
        Paths to yaml data files containing autoplotter line data

    names: str, List[str]
        Names of the simulations that correspond to the yaml data files.
        Will be placed in the legends of the plots.

    observational_data_directory: str, optional
        Optional path to the top-level observational data directory.

    file_extension: str, optional
        File extension that the figures should be referred to with.

    Returns
    -------

    auto_plotter: AutoPlotter
        An ``AutoPlotter`` instance containing metadata on e.g. observational data.

    auto_plotter_metadata: AutoPlotterMetadata
        An ``AutoPlotterMetadata`` instance, associated with ``auto_plotter``.

    line_data: Dict
        Line data loaded from the paths.
    """

    auto_plotter = AutoPlotter(
        config, observational_data_directory=observational_data_directory
    )

    file_extension = file_extension if file_extension is not None else "png"

    auto_plotter.file_extension = file_extension

    auto_plotter_metadata = AutoPlotterMetadata(auto_plotter=auto_plotter)

    line_data = load_yaml_line_data(paths=paths, names=names)

    return auto_plotter, auto_plotter_metadata, line_data


def recreate_single_figure(
    plot: VelociraptorPlot,
    line_data: Dict[str, Dict],
    output_directory: str,
    file_type: str,
) -> None:
    """
    Recreates a single figure using the data in ``line_data`` and the metadata in
    ``plot``.

    Parameters
    ----------
    plot: VelociraptorPlot
        Velociraptor plot instance (from AutoPlotter).
    line_data: Dict[str, Dict]
        Global line data, obtained from the ``load_yaml_line_data`` function.
    output_directory: str
        Output directory for the plot
    file_type: str
        Output file type (e.g. ``png``)
    """

    try:
        first_line_metadata = line_data[list(line_data.keys())[0]]["metadata"]
        fake_catalogue = FakeCatalogue(
            z=first_line_metadata["redshift"], a=first_line_metadata["scale_factor"]
        )
    except KeyError:
        fake_catalogue = FakeCatalogue()

    fig, ax = plt.subplots()

    # Add simulation data
    for line_type in valid_line_types:
        line = getattr(plot, f"{line_type}_line", None)
        if line is not None:
            for color, (name, data) in enumerate(line_data.items()):
                color_name = f"C{color}"

                try:
                    this_plot = data[plot.filename]
                    this_line_dict = this_plot["lines"][line_type]
                except KeyError:
                    continue

                if (
                    this_line_dict.get("centers", []) == []
                    and this_line_dict.get("additional_points_x", []) == []
                ):
                    # Don't plot this line, as it contains no information.
                    continue

                centers = unyt.unyt_array(this_line_dict["centers"], units=plot.x_units)
                heights = unyt.unyt_array(this_line_dict["values"], units=plot.y_units)
                errors = unyt.unyt_array(this_line_dict["scatter"], units=plot.y_units)

                ax.set_xlabel(this_plot.get("x_label", ax.get_xlabel()))
                ax.set_ylabel(this_plot.get("y_label", ax.get_ylabel()))

                # Data points from the bins with too few data points
                additional_x = unyt.unyt_array(
                    this_line_dict.get("additional_points_x", []), units=plot.x_units
                )
                additional_y = unyt.unyt_array(
                    this_line_dict.get("additional_points_y", []), units=plot.y_units
                )

                if line.scatter == "errorbar":
                    ax.errorbar(
                        centers, heights, yerr=errors, label=name, color=color_name
                    )
                elif line.scatter == "shaded":
                    ax.plot(centers, heights, label=name, color=color_name)

                    # Deal with different + and -ve errors
                    if errors.shape[0]:
                        if errors.ndim > 1:
                            down, up = errors
                        else:
                            up = errors
                            down = errors
                    else:
                        up = 0
                        down = 0

                    ax.fill_between(
                        centers,
                        heights - down,
                        heights + up,
                        color=color_name,
                        alpha=0.3,
                        linewidth=0.0,
                    )

                # line.scatter == "none":
                else:
                    ax.plot(centers, heights, label=name)

                ax.scatter(additional_x, additional_y, c=color_name)

                # Enter only if the plot has a valid Y-axis range and there are any
                # additional data points.
                if plot.y_lim is not None and len(additional_x) > 0:

                    # Draw arrows for each data point beyond X- or/and Y- axis range
                    line.highlight_data_outside_domain(
                        ax,
                        additional_x.value,
                        additional_y.value,
                        color_name,
                        (plot.x_lim[0].value, plot.x_lim[1].value),
                        (plot.y_lim[0].value, plot.y_lim[1].value),
                    )

    # Add observational data second to allow for colour precedence
    # to go to runs
    observational_data_scale_factor_bracket = [
        10 ** (log10(fake_catalogue.a) + plot.observational_data_bracket_width),
        10 ** (log10(fake_catalogue.a) - plot.observational_data_bracket_width),
    ]

    observational_data_redshift_bracket = [
        (1 - x) / x for x in observational_data_scale_factor_bracket
    ]

    valid_observational_data = load_observations(
        plot.observational_data_filenames,
        redshift_bracket=observational_data_redshift_bracket,
    )

    for index, data in enumerate(valid_observational_data, start=1):
        data.x.convert_to_units(plot.x_units)
        data.y.convert_to_units(plot.y_units)
        data.plot_on_axes(
            ax, errorbar_kwargs=dict(zorder=-10, color=f"C{index + color}")
        )

    # Finally set up metadata
    if plot.x_log:
        ax.set_xscale("log")
    if plot.y_log:
        ax.set_yscale("log")

    try:
        ax.set_xlim(*unyt.unyt_array(plot.x_lim, units=plot.x_units))
    except AttributeError:
        pass

    try:
        ax.set_ylim(*unyt.unyt_array(plot.y_lim, units=plot.y_units))
    except AttributeError:
        pass

    decorate_axes(
        ax,
        catalogue=fake_catalogue,
        comment=plot.comment,
        legend_loc=plot.legend_loc,
        redshift_loc=plot.redshift_loc,
        comment_loc=plot.comment_loc,
    )

    fig.savefig(f"{output_directory}/{plot.filename}.{file_type}")
    plt.close(fig)



import setuptools
from velociraptor import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="velociraptor",
    version=__version__,
    description="Velociraptor catalogue reading routines.",
    url="https://github.com/swiftsim/velociraptor-python",
    author="Josh Borrow",
    author_email="joshua.borrow@durham.ac.uk",
    packages=setuptools.find_packages(),
    scripts=["velociraptor-plot"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    zip_safe=False,
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
        "Operating System :: OS Independent",
    ],
    install_requires=["numpy", "unyt>=2.6.0", "h5py", "astropy"],
)
